# ruff: noqa: D104,D205,D212,D415
"""
.. include:: ../README.md
   :start-line: 1
   :end-before: </div>

.. include:: ../README.md
   :start-after: </div>
"""

from nyx_client.client import NyxClient as NyxClient
from nyx_client.configuration import BaseNyxConfig as BaseNyxConfig
from nyx_client.configuration import CohereNyxConfig as CohereNyxConfig
from nyx_client.configuration import ConfigProvider as ConfigProvider
from nyx_client.data import Data as Data
from nyx_client.utils import Parser as Parser
from nyx_client.utils import Utils as Utils
from nyx_client.utils import VectorResult as VectorResult



"""Module that manages individual Nyx Data."""

import logging
import urllib

log = logging.getLogger(__name__)


class Data:
    """Represents the data in the Nyx system.

    This class encapsulates the information and functionality related to the data
    in the Nyx system, including its metadata and content retrieval.
    """

    @property
    def name(self) -> str:
        """The unique name of the product."""
        return self._name

    @property
    def title(self) -> str:
        """The title of the product."""
        return self._title

    @property
    def creator(self) -> str:
        """The creator of the product."""
        return self._creator

    @property
    def description(self) -> str:
        """The description of the data."""
        return self._description

    @property
    def size(self) -> int:
        """The size of the product in bytes."""
        return self._size

    @property
    def content_type(self) -> str:
        """Content type (as a simple string, without IANA prefix)."""
        if self._content_type.startswith("http"):
            return self._content_type.split("/")[-1]
        return self._content_type

    @property
    def url(self):
        """The server generated url for brokered access to a subscribed dataset/product."""
        if not self._access_url:
            return self._download_url
        return self._access_url + f"?buyer_org={self._org}"

    @property
    def content(self) -> str | None:
        """The downloaded content of the product (None if not yet downloaded)."""
        return self._content

    def __init__(self, **kwargs):
        """Initialize a Data instance.

        Args:
            **kwargs: Keyword arguments containing data information.
                Required keys: 'access_url'/'download_url', 'title', 'org'

        Raises:
            KeyError: If any of the required fields are missing.
        """
        if not kwargs.get("title") or not kwargs.get("org"):
            raise KeyError(f"Required fields include 'title' and 'org'. Provided fields: {', '.join(kwargs.keys())}")
        if not (kwargs.get("access_url") or kwargs.get("download_url")):
            raise KeyError(
                f"At least one of 'access_url' or 'download_url' is required. "
                f"Provided fields: {', '.join(kwargs.keys())}"
            )

        self._title = kwargs.get("title")
        self._access_url = kwargs.get("access_url")
        self._download_url = kwargs.get("download_url")
        self._org = kwargs.get("org")
        try:
            self._size = int(kwargs.get("size"))
        except (ValueError, TypeError):
            self._size = 0
        self._creator = kwargs.get("creator")
        self._content = None
        self._name = kwargs.get("name", "unknown")
        self._description = kwargs.get("description", "unkown description")

        if content_type := kwargs.get("mediaType"):
            self._content_type = content_type
        else:
            self._content_type = "unknown"

    def __str__(self):
        return f"Data({self._title}, {self.url}, {self._content_type})"

    def download(self):
        """Download the content of the data and populate the class content field.

        This method attempts to download the content from the data's URL
        and stores it in the `content` attribute.

        Returns:
            The downloaded content, or None if the download fails.

        Note:
            If the content has already been downloaded, this method returns the cached content without re-downloading.
        """
        if self._content:
            return self._content
        url = self.url
        try:
            with urllib.request.urlopen(url) as f:
                self._content = f.read().decode("utf-8")
                return self._content
        except urllib.error.URLError as err:
            log.warning(
                "Failed to download content of data [%s], "
                "confirm the source is still available with the data producer: %s",
                self._title,
                err,
            )
            return None



import os
from getpass import getpass

import click
import requests

from nyx_client.cli.common import SDK_CLI_DEFAULT_HEADERS
from nyx_client.identity_auth import IdentityAuth


def init_env(filename: str = ".env"):
    """Interactively generate configuration and save it to the file with the given name."""
    if filename != "-":
        # Check .env exists, to ask if we should override it
        if os.path.exists(filename):
            prompt = f"'{filename}' already exists, do you wish to override it? (y/N): "
        else:
            prompt = f"Do you want to interactively create '{filename}'? (y/N): "
        # Default is always no action
        if input(prompt).lower() != "y":
            click.echo("Exiting with no changes")
            return

    # Get instance details, to get everything from API
    url = input("Enter Nyx URL: ").rstrip("/")
    email = input("Enter Nyx email: ")
    password = getpass("Enter Nyx password: ")

    headers = SDK_CLI_DEFAULT_HEADERS.copy()

    resp = requests.post(
        url + "/api/portal/auth/login",
        json={"email": email, "password": password},
        headers=headers,
    )

    # Display any failures and exit without writing
    if not resp.ok:
        raise RuntimeError(f"Unable to authorize on Nyx instance, {resp.text}")

    token = resp.json()["access_token"]
    headers["authorization"] = f"Bearer {token}"

    # Get username
    resp = requests.get(
        url + "/api/portal/users/me",
        headers=headers,
    )
    if not resp.ok:
        raise RuntimeError(f"Unable to retrieve Nyx user details, {resp.text}")

    username = resp.json()["name"]

    # Get resolver from qapi-connection
    resp = requests.get(
        url + "/api/portal/auth/qapi-connection",
        headers=headers,
    )
    if not resp.ok:
        raise RuntimeError(f"Unable to retrieve Nyx connection information, {resp.text}")

    resolver = resp.json()["resolver_url"]

    click.echo("Generating user/agent secrets")
    secrets = IdentityAuth.generate_config(resolver)
    secrets += "\n"

    # NYX creds
    secrets += f'\nNYX_USERNAME="{username}"'
    secrets += f'\nNYX_PASSWORD="{password}"'
    secrets += f'\nNYX_EMAIL="{email}"'
    secrets += f'\nNYX_URL="{url}"'

    click.echo(
        "Our high level client (NyxLangChain) expects a configured language model, "
        "however it can be configured with other language model wrappers provided by langchain "
        "providing they support 'tool calling'. "
        "Don't forget to set the appropriate API KEY as an environment variable for the language model you wish to use."
    )

    click.echo(f"Writing contents to {filename}")
    with click.open_file(filename, mode="w") as output:
        output.write(secrets)



import os

import grpc
from iotics.lib.grpc import AuthInterface, IoticsApi
from iotics.lib.grpc.helpers import KEEP_ALIVE_CHANNEL_OPTIONS

from nyx_client.configuration import BaseHostConfig
from nyx_client.identity_auth import IdentityAuth, IdentityAuthError

DID_TOKEN_DURATION = os.getenv("DID_TOKEN_DURATION", "30")


class HostClient:
    """HostClient is a context manager that provides a gRPC channel to IOTICSpace and instantiates API stubs.

    Attributes:
        verify_ssl: Whether to verify the SSL certificate.
        identity: IdentityAuth instance.
        api: IoticsApi instance.
        fqdn: Fully qualified domain name.
    """

    verify_ssl: bool
    identity: IdentityAuth
    api: IoticsApi
    fqdn: str

    def __init__(self, config: BaseHostConfig):
        """Initialize a HostClient instance.

        Args:
            config: `BaseHostConfig` instance.
        """
        host_url = config.host_url
        if host_url is None:
            host_url = os.getenv("HOST_ADDRESS", "localhost:10000")
        self.verify_ssl = config.host_verify_ssl

        try:
            self.identity = IdentityAuth(
                host_url,
                config.resolver_url,
                config.did_user_id,
                config.did_agent_id,
                config.did_agent_key_name,
                config.did_agent_name,
                config.did_agent_secret,
                DID_TOKEN_DURATION,
                self.verify_ssl,
            )
        except IdentityAuthError as error:
            raise error

        self.fqdn = self.identity.get_host_with_scheme()

    def __enter__(self):
        self.identity.refresh_token()
        self.api = IoticsApi(
            auth=self.identity,
            channel=get_channel(auth=self.identity, verify_ssl=self.verify_ssl),
        )
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.api.channel.close()


def get_channel(auth: AuthInterface, verify_ssl: bool = True, channel_options: [tuple, ...] = None) -> grpc.Channel:
    """Creates a gRPC channel to IOTICSpace and instantiates API stub.

    Args:
        auth: Required to get a space host name and authentication token.
        verify_ssl: Whether to verify the SSL certificate
        channel_options: options argument passed to the grpc.secure_channel call
          by default will use KEEP_ALIVE_CHANNEL_OPTIONS

    Returns: gRPC channel.
    """
    if channel_options is None:
        channel_options = KEEP_ALIVE_CHANNEL_OPTIONS

    channel_credentials = grpc.ssl_channel_credentials() if verify_ssl else grpc.local_channel_credentials()

    call_credentials = grpc.access_token_call_credentials(auth.get_token())
    composite_credentials = grpc.composite_channel_credentials(channel_credentials, call_credentials)
    return grpc.secure_channel(auth.get_host(), composite_credentials, options=channel_options)



"""Module for managing connection to Nyx."""

import base64
import json
import logging
import uuid
from dataclasses import dataclass
from typing import Dict, Optional

import grpc
import requests
from iotics.api.common_pb2 import Headers, Scope
from iotics.api.meta_pb2 import SparqlQueryResponse, SparqlResultType
from requests_toolbelt.multipart.encoder import MultipartEncoder

from nyx_client.configuration import BaseNyxConfig
from nyx_client.data import Data
from nyx_client.host_client import HostClient

logging.basicConfig(format="%(asctime)s %(levelname)s [%(module)s] %(message)s", level=logging.INFO)

log = logging.getLogger(__name__)

NS_IOTICS = "http://data.iotics.com/iotics#"
NS_NYX = "http://data.iotics.com/pnyx#"

DATA_NAME = NS_NYX + "productName"

_SELECT = """
        PREFIX dcat: <http://www.w3.org/ns/dcat#>
        PREFIX dct: <http://purl.org/dc/terms/>

        SELECT ?access_url ?title ?org ?name ?mediaType ?size ?creator ?description
"""

_COMMON_FILTER = f"""
          ?s <{DATA_NAME}> ?name ;
              dcat:accessURL ?access_url ;
              dct:title ?title ;
              dcat:mediaType ?mediaType ;
              dct:creator ?creator ;
              dcat:byteSize ?size ;
              dct:description ?description .
"""


@dataclass
class NyxClient:
    """A client for interacting with the Nyx system.

    This client provides methods for querying and processing data from Nyx.

    Attributes:
        config: Configuration for the Nyx client.
        host_client: Client for interacting with the host.
    """

    config: BaseNyxConfig
    host_client: HostClient

    def __init__(
        self,
        env_file: Optional[str] = None,
        config: Optional[BaseNyxConfig] = None,
    ):
        """Instantiate a new client.

        Args:
            env_file: Path to the environment file containing configuration.
            config: Pre-configured BaseNyxConfig object.
        """
        if config:
            self.config = config
        else:
            self.config = BaseNyxConfig(env_file, validate=True)

        self._token = self.config.override_token
        self._refresh = ""
        self._subscribed_data: tuple[str] = ()

        self._setup()

    def _setup(self):
        self._authorise(refresh=False)

        # Set user nickname
        self.name = self._nyx_get("users/me").get("name")
        log.debug("successful login as %s", self.name)

        self.update_subscriptions(refresh=False)

        # Get host info
        qapi = self._nyx_get("auth/qapi-connection")
        self.config.community_mode = qapi.get("community_mode", False)
        self.config.org = f"{qapi.get('org_name')}/{self.name}" if self.config.community_mode else qapi.get("org_name")
        self.config.host_config.host_url = qapi.get("grpc_url")
        self.config.host_config.resolver_url = qapi.get("resolver_url")

        self.host_client = HostClient(self.config.host_config)

    def _nyx_post(self, endpoint: str, data: dict, headers: dict = None, multipart: MultipartEncoder = None) -> dict:
        if not headers:
            headers = {"X-Requested-With": "nyx-sdk", "Content-Type": "application/json"}
        if self._token:
            headers["authorization"] = "Bearer " + self._token
        resp = requests.post(
            url=self.config.nyx_url + "/api/portal/" + endpoint,
            json=data if data else None,
            data=multipart if multipart else None,
            headers=headers,
        )
        resp.raise_for_status()

        return resp.json()

    def _nyx_get(self, endpoint: str) -> dict:
        headers = {"X-Requested-With": "nyx-sdk", "Content-Type": "application/json"}
        if self._token:
            headers["authorization"] = "Bearer " + self._token
        resp = requests.get(url=self.config.nyx_url + "/api/portal/" + endpoint, headers=headers)
        resp.raise_for_status()

        return resp.json()

    def _authorise(self, refresh=True):
        """Authorise with the configured Nyx instance using basic authorisation."""
        if not refresh and self._token:
            # If it's not fresh then we'll return and use the existing token
            return
        resp = self._nyx_post("auth/login", self.config.nyx_auth)
        log.debug("Login response: %s", resp)
        self._token = resp["access_token"]
        self._refresh = resp["refresh_token"]

    def update_subscriptions(self, refresh=True):
        """Update the list of subscribed data."""
        self._authorise(refresh)
        # Get all the data we're subscribed to, so the results are relevant to what the user wants
        purchases = self._nyx_get("purchases/transactions")
        if not purchases:
            self._subscribed_data = []
            return
        self._subscribed_data = tuple(k["product_name"] for k in purchases)

    def close(self):
        """Cleanup any resources used by the client."""
        pass

    @property
    def subscribed_data(self) -> tuple[str]:
        """Names of subscribed-to products.

        Can be explicitly updated by calling `update_subscriptions`.
        """
        return self._subscribed_data

    def _local_sparql_query(self, query: str) -> list[Dict[str, str]]:
        """Execute a SPARQL query against the configured IOTICS host.

        Args:
            query: The SPARQL query string.

        Returns:
            A list of dictionaries representing the query results.
        """
        return self._sparql_query(query, Scope.LOCAL)

    def _federated_sparql_query(self, query: str) -> list[Dict[str, str]]:
        """Execute a SPARQL query against the federated network.

        Args:
            query: The SPARQL query string.

        Returns:
            A list of dictionaries representing the query results.
        """
        return self._sparql_query(query, Scope.GLOBAL)

    def _sparql_query(self, query: str, scope: Scope) -> list[Dict[str, str]]:
        """Execute a SPARQL query and process the results.

        Args:
            query: The SPARQL query string.
            scope: The scope of the query (LOCAL or GLOBAL).

        Returns:
            A list of dictionaries representing the query results.

        Raises:
            ValueError: If the response is incomplete.
            grpc.RpcError: If there's an RPC error during the query execution.
        """
        with self.host_client as client:
            chunks: dict[int, SparqlQueryResponse] = {}
            stream = client.api.sparql_api.sparql_query(
                query,
                result_content_type=SparqlResultType.SPARQL_JSON,
                scope=scope,
                headers=Headers(transactionRef=("nyx_sdk", str(uuid.uuid4()))),
            )
            try:
                for response in stream:
                    chunks[response.payload.seqNum] = response
            except grpc.RpcError as err:
                err: grpc._channel._MultiThreadedRendezvous
                if err.code() != grpc.StatusCode.DEADLINE_EXCEEDED:
                    raise err

            sorted_chunks = sorted(chunks.values(), key=lambda r: r.payload.seqNum)
            if len(sorted_chunks) == 0:
                return []
            last_chunk = sorted_chunks[-1]

            if not last_chunk.payload.last or len(chunks) != last_chunk.payload.seqNum + 1:
                raise ValueError("Incomplete response")
            resp_json = json.loads("".join([c.payload.resultChunk.decode() for c in sorted_chunks]))
            results = []
            for binding in resp_json["results"]["bindings"]:
                inner_json = {}
                for k in binding:
                    inner_json[k] = binding[k]["value"]
                results.append(inner_json)

            return results

    def get_all_categories(self) -> list[str]:
        """Retrieve all categories from the federated network.

        Returns:
            A list of category names.
        """
        query = """
        PREFIX dcat: <http://www.w3.org/ns/dcat#>

        SELECT DISTINCT ?theme
        WHERE {
          ?s dcat:theme ?theme .
        }
        """

        return [r["theme"] for r in self._federated_sparql_query(query)]

    def get_subscribed_categories(self) -> list[str]:
        """Retrieve subscribed categories from the federated network.

        Returns:
            A list of category names.
        """
        query = f"""
        PREFIX dcat: <http://www.w3.org/ns/dcat#>

        SELECT DISTINCT ?theme
        WHERE {{
          ?s dcat:theme ?theme .
          ?s <{DATA_NAME}> ?name .
          ?s dct:creator ?creator .
          FILTER(?creator != "{self.config.org}")
          FILTER({" || ".join([f'?name = "{data}"' for data in self.subscribed_data])})
        }}
        """

        return [r["theme"] for r in self._federated_sparql_query(query)]

    def get_all_genres(self) -> list[str]:
        """Retrieve all genres from the federated network.

        Returns:
            A list of genre names.
        """
        query = """
        PREFIX dc: <http://purl.org/dc/terms/>

        SELECT DISTINCT ?genre
        WHERE {
          ?s dc:type ?genre .
        }
        """

        return [r["genre"] for r in self._federated_sparql_query(query)]

    def get_subscribed_genres(self) -> list[str]:
        """Retrieve subscribed genres from the federated network.

        Returns:
            A list of genre names.
        """
        query = f"""
        PREFIX dc: <http://purl.org/dc/terms/>

        SELECT DISTINCT ?genre
        WHERE {{
          ?s dc:type ?genre .
          ?s <{DATA_NAME}> ?name .
          ?s dct:creator ?creator .
          FILTER(?creator != "{self.config.org}")
          FILTER({" || ".join([f'?name = "{data}"' for data in self.subscribed_data])})
        }}
        """

        return [r["genre"] for r in self._federated_sparql_query(query)]

    def get_subscribed_creators(self) -> list[str]:
        """Retrieve subscribed creators from the federated network.

        Returns:
            A list of creator names.
        """
        query = f"""
        PREFIX dct: <http://purl.org/dc/terms/>

        SELECT DISTINCT ?creator
        WHERE {{
          ?s dct:creator ?creator .
          ?s <{DATA_NAME}> ?name .
          FILTER({" || ".join([f'?name = "{data}"' for data in self.subscribed_data])})
        }}
        """

        return [r["creator"] for r in self._federated_sparql_query(query)]

    def get_all_creators(self) -> list[str]:
        """Retrieve all creators from the federated network.

        Returns:
            A list of creator names.
        """
        query = """
        PREFIX dct: <http://purl.org/dc/terms/>

        SELECT DISTINCT ?creator
        WHERE {
          ?s dct:creator ?creator .
        }
        """

        return [r["creator"] for r in self._federated_sparql_query(query)]

    def get_subscribed_data(self) -> list[Data]:
        """Retrieve subscribed data from the federated network.

        Returns:
            A list of `Data` instances.
        """
        query = f"""
        {_SELECT}
        WHERE {{
          {_COMMON_FILTER}
          FILTER(?creator != "{self.config.org}")
          FILTER({" || ".join([f'?name = "{data}"' for data in self.subscribed_data])})
        }}
        """

        return [Data(**r, org=self.config.org) for r in self._federated_sparql_query(query)]

    def get_subscribed_data_for_categories(self, categories: list[str]) -> list[Data]:
        """Retrieve subscribed data for specific categories from the federated network.

        Args:
            categories: A list of category names to filter by.

        Returns:
            A list of `Data` instances matching the specified categories.
        """
        query = f"""
        {_SELECT}
        WHERE {{
          {_COMMON_FILTER}
          FILTER(?creator != "{self.config.org}")
          FILTER({" || ".join([f'?name = "{data}"' for data in self.subscribed_data])})
          ?s dcat:theme ?theme .
          FILTER({" || ".join([f'?theme = "{theme.lower()}"' for theme in categories])})
        }}
        """

        return [Data(**r, org=self.config.org) for r in self._federated_sparql_query(query)]

    def get_data_for_categories(self, categories: list[str]) -> list[Data]:
        """Retrieve all data for specific categories from the federated network.

        Args:
            categories: A list of category names to filter by.

        Returns:
            A list of `Data` instances matching the specified categories.
        """
        query = f"""
        {_SELECT}
        WHERE {{
          {_COMMON_FILTER}
          ?s dcat:theme ?theme .
          FILTER({" || ".join([f'?theme = "{theme.lower()}"' for theme in categories])})
        }}
        """

        return [Data(**r, org=self.config.org) for r in self._federated_sparql_query(query)]

    def get_subscribed_data_for_genres(self, genres: list[str]) -> list[Data]:
        """Retrieve subscribed data for specific genres from the federated network.

        Args:
            genres: A list of genre names to filter by.

        Returns:
            A list of `Data` instances matching the specified genres.
        """
        query = f"""
        {_SELECT}
        WHERE {{
          {_COMMON_FILTER}
          FILTER(?creator != "{self.config.org}")
          FILTER({" || ".join([f'?name = "{data}"' for data in self.subscribed_data])})
          ?s dct:type ?type .
          FILTER({" || ".join([f'?type = "{genre.lower()}"' for genre in genres])})
        }}
        """

        return [Data(**r, org=self.config.org) for r in self._federated_sparql_query(query)]

    def get_data_for_genres(self, genres: list[str]) -> list[Data]:
        """Retrieve data for specific genres from the federated network.

        Args:
            genres: A list of genre names to filter by.

        Returns:
            A list of `Data` instances matching the specified genres.
        """
        query = f"""
        {_SELECT}
        WHERE {{
          {_COMMON_FILTER}
          ?s dct:type ?type .
          FILTER({" || ".join([f'?type = "{genre.lower()}"' for genre in genres])})
        }}
        """

        return [Data(**r, org=self.config.org) for r in self._federated_sparql_query(query)]

    def get_data_by_name(self, name: str) -> Optional[Data]:
        """Retrieve a data based on its unique name.

        Args:
            name (str): The data unique name.

        Returns:
            The `Data` instance identified with the provided name or None if it does not exist.
        """
        query = f"""
        {_SELECT}
        WHERE {{
          {_COMMON_FILTER}
          FILTER(?name = "{name}")
        }}
        """

        data = [Data(**r, org=self.config.org) for r in self._federated_sparql_query(query)]
        # In normal state the result can either be a list of one if the data exists or 0.
        return data[0] if data else None

    def get_subscribed_data_for_creators(self, creators: list[str]) -> list[Data]:
        """Retrieve subscribed data from specific creators from the federated network.

        Args:
            creators: A list of creators to filter by.

        Returns:
            A list of `Data` instances matching the specified creators.
        """
        query = f"""
        {_SELECT}
        WHERE {{
          {_COMMON_FILTER}
          FILTER({" || ".join([f'?name = "{data}"' for data in self.subscribed_data])})
          FILTER({" || ".join([f'?creator = "{creator}"' for creator in creators])})
        }}
        """

        return [Data(**r, org=self.config.org) for r in self._federated_sparql_query(query)]

    def create_data(
        self,
        name: str,
        title: str,
        description: str,
        size: int,
        genre: str,
        categories: list[str],
        download_url: str,
        content_type: str,
        lang: str = "en",
        status: str = "published",
        preview: str = "",
        price: int = 0,
        license_url: str = "https://creativecommons.org/publicdomain/zero/1.0/",
    ) -> Data:
        """Create new data in the system.

        This method creates new data with the provided details and posts it to Nyx.

        Args:
            name: The unique identifier for the data.
            title: The display title of the data.
            description: A detailed description of the data.
            size: The size of the data, typically in bytes.
            genre: The genre or category of the data.
            categories: A list of categories the data belongs to.
            download_url: The URL where the data can be downloaded.
            content_type: The mime type of the data located at download_url.
            lang: The language of the data. Defaults to "en".
            status: The publication status of the data. Defaults to "published".
            preview: A preview or sample of the data. Defaults to an empty string.
            price: The price of the data in cents. If 0, the data is free. Defaults to 0.
            license_url: The URL of the license for the data. Defaults to Creative Commons Zero.

        Returns:
            A `Data` instance, containing the download URL and title.

        Example:
            >>> data = self.create_data(
            ...     name="uniquedataid",
            ...     title="Amazing Data",
            ...     description="This is some amazing data that does wonderful things.",
            ...     size=1024000,
            ...     genre="Software",
            ...     categories=["Utility", "Productivity"],
            ...     download_url="https://example.com/download/amazing-data",
            ...     content_type="text/csv",
            ...     price=1999  # $19.99
            ... )
            >>> print(f"Created data: {data.title}")
            Created data: Amazing Data
        """
        input_bytes = preview.encode("utf-8")
        base64_bytes = base64.b64encode(input_bytes)
        preview_base64_string = base64_bytes.decode("utf-8")

        data = {
            "name": name,
            "title": title,
            "description": description,
            "size": size,
            "genre": genre,
            "categories": categories,
            "lang": lang,
            "status": status,
            "preview": preview_base64_string,
            "downloadURL": download_url,
            "licenseURL": license_url,
            "contentType": content_type,
        }
        if price > 0:
            data["price"] = price

        multipart_data = MultipartEncoder(
            fields={
                "productMetadata": json.dumps(data),
            }
        )

        headers = {"X-Requested-With": "nyx-sdk", "Content-Type": multipart_data.content_type}

        resp = self._nyx_post("products", data, headers, multipart_data)

        args = {
            "name": name,
            "title": title,
            "description": description,
            "org": self.config.org,
            "mediaType": content_type,
            "size": size,
        }

        download_url = resp.get("downloadURL")
        access_url = resp.get("accessURL")

        if download_url:
            args["download_url"] = download_url

        if access_url:
            args["access_url"] = access_url

        return Data(**args)

    def get_data_for_creators(self, creators: list[str]) -> list[Data]:
        """Retrieve data from specific creators from the federated network.

        Args:
            creators: A list of creators to filter by.

        Returns:
            A list of `Data` instances matching the specified creators.
        """
        query = f"""
        {_SELECT}
        WHERE {{
          {_COMMON_FILTER}
          FILTER({" || ".join([f'?creator = "{creator}"' for creator in creators])})
        }}
        """
        return [Data(**r, org=self.config.org) for r in self._federated_sparql_query(query)]

    def delete_data(self, product: Data):
        """Delete the provided data from Nyx.

        Args:
            product: The data to delete.
        """
        self.delete_data_by_name(product.name)

    def delete_data_by_name(self, name: str):
        """Delete the data uniquely identified by the provided name from Nyx.

        Args:
            name: The data unique name.
        """
        headers = {"X-Requested-With": "nyx-sdk", "Content-Type": "application/json"}
        if self._token:
            headers["authorization"] = "Bearer " + self._token
        resp = requests.delete(
            url=self.config.nyx_url + f"/api/portal/products/{name}",
            headers=headers,
        )
        resp.raise_for_status()



import click

from nyx_client.cli.init_env import init_env


@click.group()
def cli():
    """NYX SDK command line utilities.

    For more information please visit https://github.com/Iotic-Labs/nyx-sdk
    """


@cli.command()
@click.argument("file", type=click.Path(writable=True, dir_okay=False), default=".env")
def init(file: str):
    """Generate an env file for the NYX client.

    Output is written to FILE (defaults to '.env' in the working directory)
    """
    init_env(file)


if __name__ == "__main__":
    cli()



# ruff: noqa: D101,D102,D107

from time import time
from typing import Optional

import names_generator
import requests
from iotics.lib.grpc.auth import AuthInterface
from iotics.lib.identity import (
    IdentityApi,
    KeyPairSecretsHelper,
    RegisteredIdentity,
    SeedMethod,
    get_rest_high_level_identity_api,
    get_rest_identity_api,
    make_identifier,
)


class IdentityAuthError(Exception):
    """Raised when `IdentityAuth` cannot be instantiated correctly."""


class IdentityAuth(AuthInterface):
    regular_api: IdentityApi
    agent: RegisteredIdentity
    host_with_scheme: str
    grpc_url: str
    token_ttl: int
    user_did: str
    user_pubkey: str
    token_last_updated: float
    host_verify_ssl: bool

    def __init__(
        self,
        space: str,
        resolver_url: Optional[str],
        user_did: str,
        agent_did: str,
        agent_key_name: str,
        agent_name: str,
        agent_secret: str,
        token_ttl: Optional[str],
        host_verify_ssl: Optional[bool],
    ):
        self.regular_api = get_rest_identity_api(resolver_url=resolver_url)
        split_url = space.partition("://")
        space = split_url[2] or split_url[0]
        self.host_verify_ssl = host_verify_ssl
        self.host_with_scheme = f"https://{space}" if host_verify_ssl else f"http://{space}"
        self.grpc_url = space
        if not resolver_url:
            index_url = f"https://{space}/index.json"
            try:
                resolver_url = requests.get(index_url).json()["resolver"]
            except requests.exceptions.ConnectionError as ex:
                raise IdentityAuthError(f"Could not fetch resolver URL from `{index_url}`.") from ex
        if token_ttl:
            self.token_ttl = int(token_ttl)
        if not agent_name.startswith("#"):
            agent_name = "#" + agent_name

        self.agent = self._get_agent(resolver_url, agent_did, agent_key_name, agent_name, agent_secret)
        self.api = get_rest_high_level_identity_api(resolver_url=resolver_url)
        self.grpc_token = self.api.create_agent_auth_token(self.agent, user_did, self.token_ttl)
        self.token_last_updated = time()
        self.user_did = user_did

        self.user_pubkey = list(self.regular_api.get_register_document(user_did).public_keys.values())[0].base58

    def get_host(self) -> str:
        return self.grpc_url

    def get_host_with_scheme(self) -> str:
        return self.host_with_scheme

    def get_token(self, ttl: Optional[int] = None) -> str:
        if self.grpc_token:
            return self.grpc_token

        self.refresh_token(ttl or self.token_ttl)
        return self.grpc_token

    def refresh_token(self, ttl: Optional[int] = None):
        self.grpc_token = self.api.create_agent_auth_token(self.agent, self.user_did, ttl or self.token_ttl)
        self.token_last_updated = time()

    @staticmethod
    def _get_agent(
        resolver_url: str,
        agent_did: str,
        agent_key_name: str,
        agent_name: str,
        agent_secret: str,
    ):
        # Gets an agent identity using the provided credentials, taking into account that the agent DID may have been
        # generated using either of two SeedMethods
        agent_secret = bytes.fromhex(agent_secret)
        api = get_rest_identity_api(resolver_url)
        agent = api.get_agent_identity(agent_secret, agent_key_name, agent_did, agent_name)
        did_from_keys = make_identifier(KeyPairSecretsHelper.get_key_pair(agent.key_pair_secrets).public_bytes)
        if did_from_keys == agent_did:
            return agent
        return api.get_agent_identity(
            agent_secret,
            agent_key_name,
            agent_did,
            agent_name,
            SeedMethod.SEED_METHOD_NONE,
        )

    @staticmethod
    def generate_config(resolver: str) -> str:
        """Create new user and agent with auth delegation and return an env-file ready string representation thereof."""
        high_level_api = get_rest_high_level_identity_api(resolver_url=resolver)
        user_seed = high_level_api.create_seed()
        agent_seed = high_level_api.create_seed()
        user_key = f"#user-{names_generator.generate_name()}"
        agent_key = f"#agent-{names_generator.generate_name()}"
        if len(user_key) > 24:
            user_key = user_key[0:23]
        if len(agent_key) > 24:
            agent_key = agent_key[0:23]

        user_deleg = "#testagent"

        if len(user_key) > 24:
            user_key = user_key[0:23]
        if len(agent_key) > 24:
            agent_key = agent_key[0:23]

        user, agent = high_level_api.create_user_and_agent_with_auth_delegation(
            user_seed=user_seed,
            user_key_name=user_key,
            agent_seed=agent_seed,
            agent_key_name=agent_key,
            delegation_name=user_deleg,
            user_name=user_key,
            agent_name=agent_key,
        )

        return f"""
#### Autogenerated by nyx_client - do not edit manually
DID_USER_DID={user.did}
DID_AGENT_DID={agent.did}
DID_AGENT_KEY_NAME="{agent_key}"
DID_AGENT_NAME="{agent_key}"
DID_AGENT_SECRET={agent_seed.hex()}
HOST_VERIFY_SSL=true # Set to false for development
"""



"""Optional module for tight integration between Nyx and LangChain."""

import logging
import os
from typing import Optional

from nyx_client.configuration import BaseNyxConfig, CohereNyxConfig, OpenAINyxConfig

try:
    from langchain_community.agent_toolkits import create_sql_agent
    from langchain_community.utilities import SQLDatabase
    from langchain_core.language_models import BaseChatModel
except ImportError as err:
    raise ImportError(
        "LangChain dependencies are not installed. "
        "Please install them with `pip install nyx_client[langchain-openai]` or "
        "`pip install nyx_client[langchain-cohere]`"
    ) from err

try:
    from langchain_openai import ChatOpenAI
except ImportError:

    class ChatOpenAI:  # noqa: D101
        def __init__(self, *args, **kwargs):  # noqa: D107
            raise ImportError(
                "LangChain OpenAI dependencies are not installed. "
                "Please install them with `pip install nyx_client[langchain-openai]`"
            )


try:
    from langchain_cohere import ChatCohere
    from langchain_cohere.sql_agent.agent import create_sql_agent as create_cohere_sql_agent
except ImportError:

    class ChatCohere:  # noqa: D101
        def __init__(self, *args, **kwargs):  # noqa: D107
            raise ImportError(
                "LangChain Cohere dependencies are not installed. "
                "Please install them with `pip install nyx_client[langchain-cohere]`"
            )


from langchain.agents.mrkl import prompt as react_prompt
from langchain_core.prompts import (
    ChatPromptTemplate,
    FewShotChatMessagePromptTemplate,
    PromptTemplate,
)

from nyx_client.client import NyxClient
from nyx_client.data import Data
from nyx_client.utils import Parser, Utils

EXAMPLES = [
    {
        "input": "List all the characters in Anna Karenina",
        "query": "SELECT DISTINCT characterNames FROM novels WHERE novelName = 'Anna Karenina'",
    },
]

SYSTEM_PREFIX = """
    You are an agent designed to interact with a SQL database.
    Given an input question, create a syntactically correct query to run, then look at the results of the query and 
    return the answer. Unless the user specifies a specific number of examples they wish to obtain.
    You can order the results by a relevant column to return the most interesting examples in the database.
    Never query for all the columns from a specific table, only ask for the relevant columns given the question.
    You have access to tools for interacting with the database.
    Only use the given tools. Only use the information returned by the tools to construct your final answer.
    You MUST double check your query before executing it. If you get an error while executing a query,
    rewrite the query and try again.

    DO NOT make any DML statements (INSERT, UPDATE, DELETE, DROP etc.) to the database.
    
    If the question does not seem related to the database, just return "I don't know" as the answer.
    
    For the following types of questions, ALWAYS respond with a list of entries from the nyx_subscriptions file:
      * What products am I subscribed to?
      * What products do I have?
      * What products are available?
      * What data do I have?
      * What data am I subscribed to?
      * What data is available to me in Nyx?
      * Any similar questions about Nyx subscriptions, products, or available data
    When responding to these queries:
      * List all entries from the nyx_subscriptions file
      * Include the name, description, and URL for each entry
      * Do not filter or limit the results for these specific queries
"""

BASIC_PREFIX = """
    Begin!

    Question: {input}
    Thought: I should look at the tables in the database to see what I can query.
    Then I should query the schema of the most relevant tables.
    {agent_scratchpad}

"""

EXAMPLE_PROMPT = ChatPromptTemplate.from_messages(
    messages=[
        ("human", "Hello, I am going to ask you a question now which you will need to answer using SQL."),
        ("ai", "OK, I will do my best to answer your question using SQL."),
        ("human", "{query}"),
    ]
)

FEW_SHOT_PROMPT = FewShotChatMessagePromptTemplate(
    examples=EXAMPLES,
    example_prompt=EXAMPLE_PROMPT,
    input_variables=["input", "agent_scratchpad"],
)

FORMAT_INSTRUCTIONS = (
    f"{react_prompt.FORMAT_INSTRUCTIONS}\n "
    f"Here are some examples of user inputs and "
    f"their corresponding SQL queries:\n"
)

TEMPLATE = "\n\n".join([SYSTEM_PREFIX, FORMAT_INSTRUCTIONS, FEW_SHOT_PROMPT.format(), BASIC_PREFIX])

DEFAULT_COHERE_MODEL = "command-r"
DEFAULT_OPENAI_MODEL = "gpt-4o-mini"


class NyxLangChain(NyxClient):
    """An opinionated client wrapping langChain to evaluate user queries against contents of a Nyx network.

    This class extends NyxClient to provide LangChain-based functionality for querying Nyx network contents.

    Note:
        The LLM must support tool calling.
    """

    def __init__(
        self,
        config: Optional[BaseNyxConfig] = None,
        env_file: Optional[str] = None,
        llm: Optional[BaseChatModel] = None,
        log_level: int = logging.WARN,
        system_prompt: Optional[str] = None,
    ):
        """Initialise a new langChain client.

        Args:
            config: Configuration for the Nyx client.
            env_file: Path to the environment file.
            llm: Language model to use.
            log_level: the logging level to use for nyx client, and langchain modules
            system_prompt: provide an override for the system prompt
        """
        super().__init__(env_file, config)
        logging.basicConfig(format="%(asctime)s %(levelname)s [%(module)s] %(message)s", level=log_level)

        # Disable langchain network requests log
        logging.getLogger("httpx").setLevel(log_level)
        logging.getLogger("httpcore").setLevel(log_level)

        self.log = logging.getLogger(__name__)
        self.log.setLevel(log_level)

        self.verbose = False
        if log_level <= logging.DEBUG:
            self.verbose = True
        self.create_agent_func = create_sql_agent

        if llm:
            self.llm = llm
            if isinstance(self.llm, ChatCohere):
                self.create_agent_func = create_cohere_sql_agent
        else:
            if isinstance(self.config, CohereNyxConfig):
                self.llm = ChatCohere(model=DEFAULT_COHERE_MODEL, cohere_api_key=self.config.api_key)
                self.create_agent_func = create_cohere_sql_agent
            elif isinstance(self.config, OpenAINyxConfig):
                self.llm = ChatOpenAI(model=DEFAULT_OPENAI_MODEL, api_key=self.config.api_key)
            else:
                raise ValueError("No language model provided and no valid config found")

        self.template = "\n\n".join([system_prompt if system_prompt else "", TEMPLATE])
        self.prompt = PromptTemplate.from_template(template=self.template)

    def query(
        self,
        query: str,
        data: Optional[list[Data]] = None,
        include_own: bool = False,
        sqlite_file: Optional[str] = None,
        update_subscribed: bool = True,
        k: int = 3,
    ) -> str:
        """Query the LLM with a user prompt and context from Nyx.

        This method takes a user prompt and invokes it against the LLM associated with this instance,
        using context from Nyx.

        Args:
            query (str): The user input.
            data (Optional[list[Data]], optional): List of products to use for context.
                If None, uses all subscribed data. Defaults to None.
            include_own (bool): Include your own data, created in Nyx, in the query.
            sqlite_file (Optional[str]): A file location to write the sql_lite file to.
            update_subscribed (bool): if set to true this will re-poll Nyx for subscribed data
            k (int): Max number of vector matches to include

        Returns:
            str: The answer from the LLM.

        Note:
            If the data list is not provided, this method updates subscriptions and retrieves all subscribed data.
        """
        if update_subscribed:
            self.update_subscriptions()
        if data is None:
            data = self.get_subscribed_data()
        if include_own:
            data.extend(self.get_data_for_creators(creators=[self.config.org]))
        self.log.debug("using products: %s", [d.title for d in data])

        parser = Parser()
        parser.data_as_vectors([d for d in data if d.content_type != "csv"], chunk_size=100)
        matching_vectors = parser.query(query, k)

        engine = Parser.data_as_db(data, matching_vectors.chunks, sqlite_file=sqlite_file, if_exists="replace")
        db = SQLDatabase(engine=engine)
        agent_executor = self.create_agent_func(self.llm, db=db, agent_type="tool-calling", verbose=self.verbose)

        res = agent_executor.invoke(
            {
                "input": self.prompt.format(
                    input=Utils.with_sources(query), agent_scratchpad="", tool_names=agent_executor.tools
                )
            }
        )

        if sqlite_file:
            os.remove(sqlite_file)

        return res["output"]



"""Module that contains utility functions, as well as tooling for manual parsing of data contained in Nyx."""

import logging
import os
import sqlite3
from io import StringIO
from typing import Any, List, Literal, Optional

import pandas as pd
from sklearn.feature_extraction.text import TfidfVectorizer
from sklearn.metrics.pairwise import cosine_similarity
from sqlalchemy import create_engine, engine
from sqlalchemy.pool import StaticPool

from nyx_client.data import Data

logging.basicConfig(format="%(asctime)s %(levelname)s [%(module)s] %(message)s", level=logging.INFO)

log = logging.getLogger(__name__)


class VectorResult:  # noqa: D101
    def __init__(self, chunks: List[str], similarities: List[float], success: bool, message: str = ""):  # noqa: D107
        self.chunks = chunks
        self.similarities = similarities
        self.success = success
        self.message = message

    def __repr__(self):
        return (
            f"VectorResult(chunks={repr(self.chunks)}, similarities={repr(self.similarities)},"
            f"success={repr(self.success)})"
        )


class Utils:
    """Utility methods for query (system) prompt modification."""

    @staticmethod
    def with_sources(prompt: str, **kwargs) -> str:
        """Expand prompt with clause to request the data sources considered to be included in the response."""
        return Utils.build_query(
            prompt
            + (
                "and also using the table nyx_subscriptions where table_name is the name of the table the "
                "information came from, retrieve the source and url of the relevant table queried."
                "When you go to say table, actually say sources. Output in markdown list format."
                "You must include sources for all requests,where the table the results came from"
                " matches the table_name in nyx_subscriptions, and the "
                "relevant url from that table"
            ),
            **kwargs,
        )

    @staticmethod
    def build_query(prompt: str, **kwargs) -> str:
        """Base prompt builder."""
        return (
            prompt
            + " "
            + (
                "Do not talk as if you are getting the results from a database, each table in the database is "
                "a file from a source. If there are no tables "
                "in the schema, or you can't see relevant ones, JUST RESPOND WITH the text 'I don't know', "
                "nothing else. \n"
                "all urls should be left in the same format, do not strip query params from the urls"
                "if you cannot find an exact match when looking for relevant tables, attempt to answer the query"
                "using the most relevant table, only provide the answer if it answers the users query."
                "Database information: \n {}".format(kwargs if kwargs else "")
            )
        )

    @staticmethod
    def with_confidence(prompt: str) -> str:
        """Like `with_sources` but also requests for a confidence score to be included in the result."""
        return prompt + (
            "Do not talk as if you are getting the results from a database, each table in the database is "
            "a file from a source. Do not make mention of any sources in your answer. If there are no tables "
            "in the schema, or you can't see relevant ones, JUST RESPOND WITH the text 'I don't know', nothing else."
            "Also, provide a confidence score between 0 and 1 for your answer. The response should be of the format: "
            '{"content": "<your response>", "confidence": <your confidence score>}'
        )


class Parser:
    """A class for processing and querying datasets from Data instances.

    This class provides methods to convert data into SQL databases or vector representations,
    and to perform queries on the processed data.

    Attributes:
        vectors: The TF-IDF vector representations of the processed content.
        vectorizer: The TfidfVectorizer instance used for creating vectors.
        chunks: The text chunks created from the processed content.
    """

    _excel_mimes = [
        # xlsx
        "vnd.openxmlformats-officedocument.spreadsheetml.sheet",
        # xls
        "vnd.ms-excel",
    ]

    def __init__(self):
        """Initialize a Parser instance."""
        self.vectors = None
        self.vectorizer = None
        self.chunks = None

    @staticmethod
    def data_as_db(
        data: List[Data],
        additional_information: List[str] = None,
        sqlite_file: Optional[str] = None,
        if_exists: Literal["fail", "replace", "append"] = "fail",
    ) -> "engine.Engine":
        """Process the content of multiple Data instances into an in-memory SQLite database.

        This method downloads the content of each Data (if it's a CSV) and converts it to an in-memory
        SQLite database. The resulting database engine is then returned for use with language models.

        Args:
            data: A list of Data instances to process.
            additional_information: List of additional information to be stored in the DB as a fallback
            sqlite_file: Provide a file for the database to reside in
            if_exists: What to do if a table already exists Defaults to "fail" can be "fail", "append", "replace"

        Returns:
            An SQLAlchemy `engine.Engine` instance for the in-memory SQLite database.

        Note:
            If the list of data is empty, an empty database engine is returned.
        """
        connection_str = ":memory:"
        if sqlite_file:
            # Ensure the directory exists
            os.makedirs(os.path.dirname(sqlite_file), exist_ok=True)
            connection_str = sqlite_file

        connection = sqlite3.connect(connection_str, check_same_thread=False)
        db_engine = create_engine(
            "sqlite://",
            creator=lambda: connection,
            poolclass=StaticPool,
            connect_args={"check_same_thread": False},
        )

        if len(data) == 0:
            return db_engine
        tables = []
        for d in data:
            content = d.download()
            if content is None:
                log.debug("Not adding table for %s as no content was found", d.title)
                continue
            try:
                if d.content_type == "csv":
                    data = pd.read_csv(StringIO(content))
                elif d.content_type in Parser._excel_mimes:
                    data = pd.read_excel(StringIO(content))
                elif d.content_type == "json":
                    data = pd.read_json(StringIO(content))
                else:
                    log.warning("%s is unsupported type %s", d.title, d.content_type)
                    continue
                data.columns = Parser.normalise_values(data.columns)
                table_name = Parser.normalise_values([d.title])[0]
                data.to_sql(table_name, db_engine, index=False)
            except pd.errors.ParserError:
                if d.content_type in ["csv", "json", *Parser._excel_mimes]:
                    log.warning("%s could not be processed as a %s", d.title, d.content_type)
                    continue
            except Exception as e:
                log.warning("Unexpected error for %s: %s", d.title, e)
                continue
            tables.append([d.title, d.url, d.title.replace(" ", "_"), d.description])
        if len(tables) > 0:
            df = pd.DataFrame(tables)
            df.columns = ["file_title", "url", "table_name", "description"]
            df.to_sql("nyx_subscriptions", db_engine, index=False, if_exists=if_exists)

        if additional_information:
            df = pd.DataFrame(additional_information)
            df.columns = ["context"]
            df.to_sql("additional_information", db_engine, index=False, if_exists=if_exists)
        return db_engine

    @staticmethod
    def normalise_values(values: List[str]) -> List[str]:
        """Normalise names in a list of values.

        Args:
            values (List[str]): A list of values to normalise.

        Returns:
            List[str]: A list of normalised values.
        """
        return [
            value.lower().replace(" ", "_").replace(".", "_").replace("-", "_").replace("(", "_").replace(")", "_")
            for value in values
            if value
        ]

    def data_as_vectors(self, data: List[Data], chunk_size: int = 1000):
        """Process the content of multiple Data instances into vector representations.

        This method downloads the content of each Data, combines it, chunks it,
        and creates a TF-IDF vectorizer for the chunks.

        Args:
            data (List[Data]): A list of Data instances to process.
            chunk_size (int, optional): The size of each chunk when splitting the content. Defaults to 1000.

        Returns:
            Parser: The current Parser instance with updated vectors, vectorizer, and chunks.

        Note:
            If no content is found in any of the data, the method returns without processing.
        """
        contents = ""
        for d in data:
            content = d.download()
            if content:
                contents += content
        if contents == "":
            return

        self.chunks = self._chunk_text(contents, chunk_size)
        if len(self.chunks) == 0:
            return

        self.vectorizer = TfidfVectorizer()
        self.vectors = self.vectorizer.fit_transform(self.chunks)

        return self

    @staticmethod
    def _chunk_text(text: str, chunk_size: int) -> List[str]:
        """Split a text into chunks of a specified size.

        Args:
            text (str): The text to be chunked.
            chunk_size (int): The maximum number of words in each chunk.

        Returns:
            List[str]: A list of text chunks.
        """
        words = text.split()
        return [" ".join(words[i : i + chunk_size]) for i in range(0, len(words), chunk_size)]

    def find_matching_chunk(self, query_vector: Any, k: int = 3) -> VectorResult:
        """Find the best matching chunks for a given query vector.

        Args:
            query_vector (Any): The vector representation of the query.
            k (int): The number of top matches to return. Defaults to 3.

        Returns:
            VectorResult: An object containing the best matching chunks, their similarity scores, and success status.

        Raises:
            ValueError: If the content has not been processed yet.
        """
        if self.vectors is None:
            return VectorResult(
                chunks=[],
                similarities=[],
                success=False,
                message="content has either not been processed, or you have no plain text",
            )

        similarities = cosine_similarity(query_vector, self.vectors)
        top_indices = similarities[0].argsort()[-k:][::-1]  # Get indices of top k similarities

        top_chunks = [self.chunks[i] for i in top_indices]
        top_similarities = similarities[0][top_indices].tolist()

        return VectorResult(
            chunks=top_chunks,
            similarities=top_similarities,
            success=True,
        )

    def query(self, text: str, k: int = 3) -> VectorResult:
        """Query the processed content with a text string.

        Args:
            text (str): The query text.
            k (int): The number of top matches to return. Defaults to 3.

        Returns:
            VectorResult: An object containing the best matching chunks and related information,
                          or an error message if the content has not been processed.
        """
        if self.vectorizer is None:
            return VectorResult(
                chunks=[],
                similarities=[],
                success=False,
                message="content not processed, or not present on the data (do you have access?)",
            )
        query_vector = self.vectorizer.transform([text])
        return self.find_matching_chunk(query_vector, k)



# ruff: noqa: D102

"""SDK configuration classes."""

import enum
import inspect
import json
import os
from dataclasses import dataclass
from typing import Any, Dict, Optional

from dotenv import dotenv_values


@dataclass
class BaseHostConfig:
    """Configuration for the Host client.

    Attributes:
        host_url (str): The URL of the host.
        host_verify_ssl (bool): Whether to verify the SSL certificate of the host.
        resolver_url (str): The URL of the resolver.
        did_user_id (str): The DID of the user.
        did_agent_id (str): The DID of the agent.
        did_agent_key_name (str): The key name of the agent.
        did_agent_name (str): The name of the agent.
        did_agent_secret (str): The secret of the agent.
    """

    required = [
        "did_user_id",
        "did_agent_id",
        "did_agent_key_name",
        "did_agent_name",
        "did_agent_secret",
    ]

    def __init__(self, config: dict[str, Any]):
        """Instantiates a new configuration.

        Args:
            config: A dictionary containing the configuration parameters.

        Raises:
            RuntimeError: If the configuration is invalid
        """
        self._host_url: str = config.get("HOST_URL")
        self._host_verify_ssl: bool = config.get("HOST_VERIFY_SSL", True)
        self._resolver_url: str = config.get("RESOLVER_URL")
        self._did_user_id: str = config.get("DID_USER_DID")
        self._did_agent_id: str = config.get("DID_AGENT_DID")
        self._did_agent_key_name: str = config.get("DID_AGENT_KEY_NAME")
        self._did_agent_name: str = config.get("DID_AGENT_NAME")
        self._did_agent_secret: str = config.get("DID_AGENT_SECRET")

        self._validate()

    def _validate(self):
        missing = [k for k in BaseHostConfig.required if getattr(self, k) is None]

        if missing:
            raise RuntimeError(f"Missing Required env vars {missing}")

    @property
    def host_url(self) -> str:
        return self._host_url

    @host_url.setter
    def host_url(self, url):
        self._host_url = url

    @property
    def host_verify_ssl(self) -> bool:
        return self._host_verify_ssl

    @host_verify_ssl.setter
    def host_verify_ssl(self, verify):
        self._host_verify_ssl = verify

    @property
    def resolver_url(self) -> str:
        return self._resolver_url

    @resolver_url.setter
    def resolver_url(self, url):
        self._resolver_url = url

    @property
    def did_user_id(self) -> str:
        return self._did_user_id

    @did_user_id.setter
    def did_user_id(self, did):
        self._did_user_id = did

    @property
    def did_agent_id(self) -> str:
        return self._did_agent_id

    @did_agent_id.setter
    def did_agent_id(self, did):
        self._did_agent_id = did

    @property
    def did_agent_key_name(self) -> str:
        return self._did_agent_key_name

    @did_agent_key_name.setter
    def did_agent_key_name(self, key_name):
        self._did_agent_key_name = key_name

    @property
    def did_agent_name(self) -> str:
        return self._did_agent_name

    @did_agent_name.setter
    def did_agent_name(self, agent_name):
        self._did_agent_name = agent_name

    @property
    def did_agent_secret(self) -> str:
        return self._did_agent_secret

    @did_agent_secret.setter
    def did_agent_secret(self, secret):
        self._did_agent_secret = secret


@dataclass
class BaseNyxConfig:
    """Configuration for the Nyx client.

    Attributes:
        host_config (BaseHostConfig): Configuration for the host client.
        nyx_url (str): The URL of the Nyx instance.
        nyx_username (str): The username of the Nyx user.
        nyx_email (str): The email of the Nyx user.
        nyx_password (str): The password of the Nyx user.
        org (str): The organisation name.
        community_mode (bool): Whether the host is in community mode.
    """

    required_if_token_empty = [
        "nyx_username",
        "nyx_email",
        "nyx_password",
    ]

    def __init__(
        self,
        env_file: Optional[str] = None,
        override_token: str = "",
        validate: bool = True,
        host_config: Optional[BaseHostConfig] = None,
    ):
        """Instantiate a new nyx base configuration.

        Args:
            env_file: Path to the environment file. If None, dotenv will search for a .env file.
            override_token: Token to override the default authentication. Defaults to username and password.
            validate: Whether to validate the environment variables. Defaults to False.
            host_config: Configuration for the host client.

        """
        # Load from .env, note env vars will not be overwritten
        # If no env file supplied dotenv will traverse up the directory tree looking for a .env file
        vals: Dict[str, Optional[str]] = dotenv_values(dotenv_path=env_file if env_file else None)

        self._override_token: str = override_token

        self._nyx_url: str = vals.get("NYX_URL", "https://nyx-community-1.dev.iotics.space")
        self._nyx_username: str = vals.get("NYX_USERNAME")
        self._nyx_email: str = vals.get("NYX_EMAIL")
        self._nyx_password: str = vals.get("NYX_PASSWORD")
        self._org: str = ""
        self._host_url: str = ""
        self._host_verify_ssl: bool = vals.get("HOST_VERIFY_SSL", "True").lower() == "true"
        self._resolver_url: str = ""
        self._community_mode: bool = False
        self._validated = False

        if validate and not self._validated:
            self.validate()

        self._host_config = host_config if host_config else BaseHostConfig(vals)

    def validate(self):
        if self._validated:
            return

        # Only check for auth if no token is supplied
        if not self.override_token:
            missing = [k for k in BaseNyxConfig.required_if_token_empty if getattr(self, k) is None]
            if len(missing) > 0:
                raise OSError(f"Missing Required env vars {missing}")

    def __str__(self):
        return json.dumps(self.__dict__)

    @property
    def nyx_url(self) -> str:
        return self._nyx_url

    @nyx_url.setter
    def nyx_url(self, url):
        self._nyx_url = url

    @property
    def nyx_username(self) -> str:
        return self._nyx_username

    @property
    def nyx_email(self) -> str:
        return self._nyx_email

    @property
    def nyx_password(self) -> str:
        return self._nyx_password

    @property
    def nyx_auth(self) -> dict:
        return {"email": self.nyx_email, "password": self.nyx_password}

    @property
    def org(self) -> str:
        return self._org

    @org.setter
    def org(self, org):
        self._org = org

    @property
    def community_mode(self) -> bool:
        return self._community_mode

    @community_mode.setter
    def community_mode(self, mode):
        self._community_mode = mode

    @property
    def override_token(self) -> str:
        return self._override_token

    @override_token.setter
    def override_token(self, token):
        self._override_token = token

    @property
    def host_config(self) -> BaseHostConfig:
        return self._host_config

    @host_config.setter
    def host_config(self, config):
        self._host_config = config


@dataclass
class CohereNyxConfig(BaseNyxConfig):
    """Cohere specific configuration for the Nyx client."""

    def __init__(
        self,
        api_key: Optional[str] = None,
        env_file: Optional[str] = None,
        override_token: str = "",
        validate: bool = True,
        host_config: Optional[BaseHostConfig] = None,
    ):
        """Instantiate a new Cohere nyx configuration.

        Args:
            api_key: The Cohere API key.
            env_file: Path to the environment file. If None, dotenv will search for a .env file.
            override_token: Token to override the default authentication. Defaults to username and password.
            validate: Whether to validate the environment variables. Defaults to False.
            host_config: Configuration for the host client.
        """
        if api_key:
            self._api_key: str = api_key
        else:
            self._api_key: str = os.getenv("COHERE_API_KEY", "")
        super().__init__(env_file, override_token, validate, host_config)

        if validate:
            self.validate()

    def validate(self):
        if not self._api_key:
            raise OSError("Missing Required env var: COHERE_API_KEY")
        if not self._validated:
            super().validate()

    @property
    def api_key(self) -> str:
        return self._api_key

    @api_key.setter
    def api_key(self, api_key: str):
        self._api_key = api_key


@dataclass
class OpenAINyxConfig(BaseNyxConfig):
    """OpenAI specific configuration for the Nyx client."""

    def __init__(
        self,
        api_key: Optional[str] = None,
        env_file: Optional[str] = None,
        override_token: str = "",
        validate: bool = True,
        host_config: Optional[BaseHostConfig] = None,
    ):
        """Instantiate a new OpenAI nyx configuration.

        Args:
            api_key: The OpenAI API key.
            env_file: Path to the environment file. If None, dotenv will search for a .env file.
            override_token: Token to override the default authentication. Defaults to username and password.
            validate: Whether to validate the environment variables. Defaults to False.
            host_config: Configuration for the host client.
        """
        if api_key:
            self._api_key: str = api_key
        else:
            self._api_key: str = os.getenv("OPENAI_API_KEY", "")
        super().__init__(env_file, override_token, validate, host_config)

        if validate:
            self.validate()

    def validate(self):
        if not self._api_key:
            raise OSError("Missing Required env var: OPENAI_API_KEY")
        if not self._validated:
            super().validate()

    @property
    def api_key(self) -> str:
        return self._api_key

    @api_key.setter
    def api_key(self, api_key: str):
        self._api_key = api_key


@enum.unique
class ConfigType(str, enum.Enum):
    """nyx configuration types."""

    BASE = "base"
    OPENAI = "openai"
    COHERE = "cohere"


class ConfigProvider:
    """Factory class to create Nyx configuration objects.

    Attributes:
        config_classes (dict): A dictionary of configuration classes: base, openai, cohere.
    """

    config_classes = {
        ConfigType.BASE: BaseNyxConfig,
        ConfigType.OPENAI: OpenAINyxConfig,
        ConfigType.COHERE: CohereNyxConfig,
    }

    @staticmethod
    def create_config(
        config_type: ConfigType,
        api_key: Optional[str] = None,
        env_file: Optional[str] = None,
        override_token: str = "",
        validate: bool = True,
        host_config: Optional[BaseHostConfig] = None,
    ):
        config_class = ConfigProvider.config_classes.get(config_type)
        if not config_class:
            raise ValueError(f"Unknown config type: {config_type}")

        constructor_signature = inspect.signature(config_class.__init__)
        parameters = constructor_signature.parameters

        # Dynamically create arguments to pass based on what the constructor expects
        kwargs = {
            "env_file": env_file,
            "override_token": override_token,
            "validate": validate,
            "host_config": host_config,
        }

        # If 'api_key' is part of the constructor, include it in the arguments
        if "api_key" in parameters:
            kwargs["api_key"] = api_key

        # Call the config class with the correct arguments
        return config_class(**kwargs)
